summaryrefslogtreecommitdiffstats
path: root/tools/cuda-ab/vecadd_loop.cu
blob: bde3d058d3eddc60e81fe01afc42fa4264e66800 (plain)
1
2
3
4
5
6
7
8
9
10
11
12
13
14
15
16
17
18
19
20
21
22
23
24
25
26
27
28
29
30
31
32
33
34
35
36
37
38
39
40
41
42
43
44
45
46
47
48
49
50
51
52
53
54
55
56
57
58
59
60
61
62
63
64
65
66
67
68
69
70
71
72
73
74
75
76
77
78
79
80
81
82
83
84
85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
108
109
110
111
112
113
114
115
116
117
118
119
120
121
// SPDX-License-Identifier: BSD-3-Clause
/*
 * vecadd_loop.cu - 100 repeated vectorAdd launches, soak-style.
 *
 * The same three device buffers are reused for 100 identical vadd launches:
 * the point is repetition and stability, not new data. Every 10th launch the
 * result is copied back and verified element by element against the
 * host-computed a+b, so a silent corruption that only appears after many
 * launches is caught.
 *
 * Output lines are stable and greppable:
 *
 *   DEVICE <name> cc=<major>.<minor> sm=<count> mem=<bytes>
 *   VERIFY launch=<n> bad=<mismatches>
 *   LOOP launches=100 bad=<mismatches>
 *   STATUS OK|FAIL
 *
 * Build with a static CUDA runtime (--cudart=static) so the guest only needs
 * the driver's libcuda.so, not a CUDA toolkit installation:
 *
 *   nvcc -O2 --cudart=static -o vecadd_loop vecadd_loop.cu
 *
 * Then, from the guest:  /tmp/vecadd_loop
 */

#include <cstdio>
#include <cstdlib>
#include <cstdint>
#include <cuda_runtime.h>

#define NLAUNCH 100
#define VERIFY_EVERY 10

#define CHECK(x)                                                         \
	do {                                                             \
		cudaError_t e_ = (x);                                    \
		if (e_ != cudaSuccess) {                                 \
			printf("CUDA_ERROR %s at line %d\n",             \
			       cudaGetErrorString(e_), __LINE__);        \
			printf("STATUS FAIL\n");                         \
			return 1;                                        \
		}                                                        \
	} while (0)

__global__ void vadd(const float *a, const float *b, float *c, int n)
{
	int i = blockIdx.x * blockDim.x + threadIdx.x;

	if (i < n)
		c[i] = a[i] + b[i];
}

int main(void)
{
	cudaDeviceProp prop;
	const int n = 1 << 15;
	size_t bytes = (size_t)n * sizeof(float);
	float *h_a, *h_b, *h_c;
	float *d_a, *d_b, *d_c;
	int bad = 0, i, launch;

	setvbuf(stdout, NULL, _IONBF, 0);

	CHECK(cudaGetDeviceProperties(&prop, 0));
	printf("DEVICE %s cc=%d.%d sm=%d mem=%llu\n", prop.name,
	       prop.major, prop.minor, prop.multiProcessorCount,
	       (unsigned long long)prop.totalGlobalMem);

	h_a = (float *)malloc(bytes);
	h_b = (float *)malloc(bytes);
	h_c = (float *)malloc(bytes);
	if (!h_a || !h_b || !h_c)
		return 1;

	for (i = 0; i < n; i++) {
		h_a[i] = (float)(i % 1000) * 0.5f;
		h_b[i] = (float)(i % 997) * 0.25f;
	}

	CHECK(cudaMalloc(&d_a, bytes));
	CHECK(cudaMalloc(&d_b, bytes));
	CHECK(cudaMalloc(&d_c, bytes));
	CHECK(cudaMemcpy(d_a, h_a, bytes, cudaMemcpyHostToDevice));
	CHECK(cudaMemcpy(d_b, h_b, bytes, cudaMemcpyHostToDevice));

	for (launch = 1; launch <= NLAUNCH; launch++) {
		vadd<<<(n + 255) / 256, 256>>>(d_a, d_b, d_c, n);
		CHECK(cudaGetLastError());
		if (launch % VERIFY_EVERY != 0)
			continue;
		CHECK(cudaDeviceSynchronize());
		CHECK(cudaMemcpy(h_c, d_c, bytes, cudaMemcpyDeviceToHost));
		{
			int vbad = 0;

			for (i = 0; i < n; i++) {
				if (h_c[i] != h_a[i] + h_b[i])
					vbad++;
			}
			printf("VERIFY launch=%d bad=%d\n", launch, vbad);
			bad += vbad;
		}
	}
	CHECK(cudaDeviceSynchronize());

	printf("LOOP launches=%d bad=%d\n", NLAUNCH, bad);

	if (bad != 0) {
		printf("STATUS FAIL\n");
		return 1;
	}

	printf("STATUS OK\n");
	CHECK(cudaFree(d_a));
	CHECK(cudaFree(d_b));
	CHECK(cudaFree(d_c));
	free(h_a);
	free(h_b);
	free(h_c);
	return 0;
}